#!/usr/bin/env python
#
# Copyright (c) 2004, 2005 Nokia
# Authors: Veli Mankinen <veli.mankinen@movial.fi>
#          Timo Savola <tsavola@movial.fi>
#
# Licensed under GPL, see COPYING for details.
#

import os, sys, socket

sys.path += [os.path.dirname(os.path.realpath(sys.argv[0])) + "/../lib/python2.3"]

from sb import config, sbrsh
import dialog

class MainMenu (Exception):
	pass

def handle_exit_code(code):
	if code in (d.DIALOG_CANCEL, d.DIALOG_ESC):
		if d.yesno("Do you want to return to main menu?", width=39, height=5) == d.DIALOG_OK:
			raise MainMenu()
		return False
	return True

def widget_inputbox(win_title, question, init_value):
	while True:
		code, answer = d.inputbox(question, height=8, width=60,
					  init=init_value, title=win_title)
		if handle_exit_code(code):
			break
	return answer

def widget_menu(win_title, items, prompt = "", w = 60, h = 16, mh = 10):
	while True:
		code, tag = d.menu(prompt, title=win_title, width=w, height=h,
				   menu_height=mh, choices=items)
		if handle_exit_code(code):
			break	
	return tag

def get_size(msg, min_width):
	lines = msg.split("\n")

	width = min_width
	for line in lines:
		width = max(len(line), width)

	width = min(width, 72)

	height = 0
	for line in lines:
		l = len(line)
		if l > 0:
			height = height + 1 + (l - 1) / width
		else:
			height = height + 1

	height = min(height, 14)

	return width, height

def widget_yesno(question, min_width = 0):
	width, height = get_size(question, min_width)

	code = d.yesno(question,
		       width  = width  + 4,
		       height = height + 4)

	return code == d.DIALOG_OK

def widget_scrollbox(message, title = "Scratchbox", min_width = 0):
	width, height = get_size(message, min_width)

	code = d.scrollbox(message,
			   width  = width  + 4,
			   height = height + 4,
			   title  = title)

	return code == d.DIALOG_OK

def choose_target(title):
	targets = []
	for name in config.list_targets():
		info = config.get_target_info(name)
		if info:
			targets.append([name, info.compiler])
	if targets:
		return widget_menu(title, targets)
	else:
		widget_scrollbox("\nNo targets found. Please create a target first.\n", title="Error")
		raise MainMenu()

##
## Setup

def setup(selected = None):
	if not selected:
		targets = [
			("<NEW>", "Create a new target"),
			("-----", "-------------------")]
		for name in config.list_targets():
			info = config.get_target_info(name)
			if info:
				targets.append([name, info.compiler])

		while True:
			selected = widget_menu("Install into target", targets)
			if selected == "<NEW>":
				name = widget_inputbox("Target name",
						       "Enter name for the new target", "")
				if name not in config.list_targets():
					break
				if widget_yesno("Target " + name + " already exists!\n\n" + \
						"Do you wish to overwrite the configuration?"):
					break
			elif selected != "-----":
				name = selected
				break

	compiler  = setup_compiler()
	devkits   = setup_devkits()
	cputransp = setup_cputransp(devkits)

	config.setup(name, compiler, devkits, cputransp, True)

	if cputransp == "sbrsh" and \
	   widget_yesno("Do you wish to configure sbrsh for the target?\n" + \
			"(You can also do it with the 'sbrsh-conf' command.)"):
		host = widget_inputbox("Sbrsh configuration",
				       "Enter hostname and port (optional) of the target device\n" + \
				       "Syntax: <host>[:<port>]", "")

		nfs_host = widget_inputbox("Sbrsh configuration",
					   "Enter hostname of the NFS server\n",
					   socket.gethostbyname_ex(socket.gethostname())[0])

		sbrsh.configure(name, host, nfs_host)

		i = host.find(":")
		if i >= 0:
			hostname = host[:i]
		else:
			hostname = host

		if widget_yesno("Do you want to add yourself to the access list of the sbrshd" + \
				" running at the target device?\n\n" + \
				"(This requires that %s has ssh daemon running" % hostname + \
				" and sbrshd already installed, and that you have root access.)"):
			sbrsh.add(hostname)

	if compiler != "host-gcc":
		inst_rootstrap = widget_yesno("Do you wish to extract a rootstrap on the target?")
		if inst_rootstrap:
			rootstrap(name)
	else:
		inst_rootstrap = False

	if not inst_rootstrap:
		if widget_yesno("Do you wish to install files to the target?"):
			install(name)

	if widget_yesno("Do you wish to select the target?"):
		select(name)

def setup_compiler():
	compiler_list = []

	if len(config.list_compilers()) == 0:
		widget_scrollbox("\nNo compilers found. Please install a compiler first.\n", title="Error")
		raise MainMenu()

	for name in config.list_compilers():
		info = "cross"
		if name == "host-gcc":
			info = "host"
		compiler_list.append((name, info))
	
	return widget_menu("Select compiler", compiler_list)

def setup_devkits():
	avail = []
	for name in config.list_devkits():
		deps = config.get_devkit_deps(name)
		if deps:
			info = "requires: " + config.flatten(deps, ", ")
		else:
			info = ""

		avail.append((name, info))

	selected = []
	message = ""

	while True:
		remain = [
			("DONE",  "Done selecting"),
			("CLEAR", "Clear selected"),
			("-----", "--------------") ]
		for pair in avail:
			if pair[0] not in selected:
				remain.append(pair)

		name = widget_menu("Select devkits", remain, "Selected devkits: %s\n\n%s" %
				   (config.flatten(selected, ", "), message), h=20)
		
		message = ""

		if name == 'DONE':
			break 

		elif name == 'CLEAR':
			selected = []

		elif name != '-----':
			try:
				new_selected = selected + [name]
				selected = config.normalize_devkits(new_selected)
			except Error, e:
				message = str(e)
			
	return selected

def setup_cputransp(devkits):
	cputransp_list = []
	for method in config.list_cputransps(devkits):
		if method.startswith("qemu"):
			cputransp_list.append((method, "Emulation"))
		elif method.startswith("sbrsh"):
			cputransp_list.append((method, "Remote execution"))
		else:
			cputransp_list.append((method, ""))
	cputransp_list.append(("none", ""))

	method = widget_menu("Select CPU-transparency method", cputransp_list)
	return config.processable(method, "")

##
## Install

def install(selected = None, default = 1):
	if not selected:
		selected = choose_target("Install into target")

	info = config.get_target_info(selected)
	if info and info.compiler == "host-gcc":
		choices = [
			("/etc",      "Install passwd, group, mtab, etc.", 0),
			("Devkits",   "Install files needed by devkits", 0) ]
	else:
		choices = [
			("C-library",   "Install libc and related binaries", default),
			("C-runtime",   "Install c runtime without libc", 0),
			("/etc",        "Install passwd, group, mtab, etc.", default),
			("Devkits",     "Install files needed by devkits", default),
			("Debug links", "Install debugging symlinks", default)]

		for name, needed in [
				('fakeroot', '1.4'),
				('gdb', None),
				('strace', None) ]:
			list = config.find_device_tool_versions(name, info.compiler)
			if list:
				version, path = list[-1]
				desc = 'Install %s %s' % (name, version)

				if needed and config.parse_version(version) >= config.parse_version(needed):
					tool_default = default
				else:
					tool_default = 0

				choices.append((name, desc, tool_default))

		if not config.get_compiler_info(info.compiler).prefixedInstall and selected!=config.current(None):
			if widget_yesno("It looks like you are using a toolchain\n"
							+"that doesn't support installations to nonactive target\n"
							+"Do you want to select target first?"):
				widget_scrollbox("OK, run sb-menu again when you get to the prompt and select Install.")
				select(selected)
			else:
				del choices[0]


	while True:
		code, tags = d.checklist(
			title="Install files to target",
			text="",
			height=15, width=60, list_height=8,
			choices=choices)

		if handle_exit_code(code):
			break

	tools = []
	for name in "fakeroot", "gdb", "strace":
		if name in tags:
			tools.append(name)

	config.install(selected, "C-library" in tags, "C-runtime" in tags, "/etc" in tags, "Devkits" in tags, "Debug links" in tags, tools)

##
## Rootstrap

def rootstrap(selected = None):
	if not selected:
		selected = choose_target("Extract rootstrap into target")

	choices = [
		("File",     "Select a local file"),
		("Download", "Download a file via HTTP or FTP") ]
	choice = widget_menu("Select action", choices)

	if choice == "File":
		path = os.path.join(config.scratchbox_dir, "packages") + "/"

		while True:
			code, path = d.fselect(path, 10, 60, title="Use file as rootstrap")

			if handle_exit_code(code):
				if os.path.isdir(path):
					widget_scrollbox(path + " is a directory",
							 title="Error")
				else:
					break

		config.rootstrap(selected, path=path)

	elif choice == "Download":
		url = ""

		while True:
			url = widget_inputbox("Give URL to the file", "", url)
			if not url:
				continue

			try:
				config.rootstrap(selected, url=url)
				break

			except config.DownloadError, e:
				widget_scrollbox(str(e).replace(": ", ":\n"),
						 title="Error")

	if widget_yesno("Do you wish to install files to the target?"):
		install(selected, default=0)

##
## Select

def select(selected = None):
	if not selected:
		selected = choose_target("Select target")

	try:
		d.clear()
		config.select(selected)

	except config.SessionError, e:
		widget_scrollbox(str(e), title="Error")

##
## Reset

def reset(selected = None):
	if not selected:
		selected = choose_target("Reset target")

	if widget_yesno("Are you sure you want to delete all files of target %s?" % selected):
		config.reset(selected, True)

		if selected == config.current() and \
		   widget_yesno("Do you wish to reselect the target?"):
			select(selected)

##
## Remove

def remove(selected = None):
	if not selected:
		selected = choose_target("Remove target")

	if selected == config.current():
		widget_scrollbox("You cannot remove the active target!\n" + \
				 "Select another target first.",
				 title="Error")
	elif widget_yesno("Are you sure you want to delete the configuration and all files of " + \
			  "target %s?" % selected):
		config.remove(selected, True)

##
## Show

def show(selected = None):
	if not selected:
		selected = choose_target("Show target")

	target = config.get_target_info(selected)
	comp = config.get_compiler_info(target.compiler)

	msg = ("Compiler:         %s\n" % target.compiler) + \
	      ("Architecture:     %s\n" % comp.arch) + \
	      ("Sub-architecture: %s\n" % comp.subarch) + \
	      ("C-library:        %s\n" % comp.clib) + \
	       "\n" + \
	      ("Devkits:          %s\n" % config.printable(config.flatten(target.devkits, " "))) + \
	      ("CPU-transparency: %s\n" % config.printable(target.cputransp))

	widget_scrollbox(msg, title=selected, min_width=32)

##
## Killall

def killall(selected = None):
	config.killall(
		int(widget_inputbox("Signal type","Enter kill signal", "15"))
		)

##
## Action

# use scratchbox's dialog is host os doesn't seem to have one
if ( os.environ.has_key('PATH') ) :
	os.environ['PATH'] += ':' + os.path.split(os.path.realpath(sys.argv[0]))[0]
else :
	os.environ['PATH'] = os.path.split(os.path.realpath(sys.argv[0]))[0]

d = dialog.Dialog(dialog="dialog")
d.add_persistent_args(["--backtitle", "Scratchbox configuration"])

choices = [
	("Exit",      ""),
	("Setup",     "Setup a target"),
	("Install",   "Install files to a target"),
	("Rootstrap", "Extract rootstraps"),
	("Select",    "Activate a target"),
	("Reset",     "Reset a target"),
	("Remove",    "Remove a target"),
	("Show",      "Show target information"),
	("Killall",   "Kill all processes in scratchbox"),
	]

while True:
	try:
		action = widget_menu("Select action", choices, h=16, mh=10)

		if action == "Exit":
			break

		eval(action.lower() + "()")

	except MainMenu:
		pass

	except dialog.error, e:
		widget_scrollbox(e.complete_message(), title="Error")

	except config.Error, e:
		widget_scrollbox(str(e).replace(": ", ":\n"), title="Error")

	except KeyboardInterrupt:
		sys.exit(1)

	except Exception, e:
		widget_scrollbox(str(e), title="Error")
