#
# Copyright (c) 2004, 2005 Nokia
# Author: Timo Savola <tsavola@movial.fi>
#
# Licensed under GPL, see COPYING for details.
#

import sys, os, shutil, signal, grp, re
from os.path import join, exists, isdir, isfile, islink, basename, dirname, realpath

if 'set' not in __builtins__:
	from sets import Set as set

tools_dir = dirname(realpath(sys.argv[0]))
if tools_dir.endswith("/tools/bin"):
	scratchbox_dir = tools_dir[:-10]
else:
	scratchbox_dir = "/scratchbox"

users_dir       = join(scratchbox_dir, "users")
tools_dir       = join(scratchbox_dir, "tools", "bin")
etc_dir         = join(scratchbox_dir, "etc")
compilers_dir   = join(scratchbox_dir, "compilers")
devkits_dir     = join(scratchbox_dir, "devkits")
devicetools_dir = join(scratchbox_dir, "device_tools")

if isdir("/targets"):
	sandbox_dir = "/"
else:
	sandbox_dir = join(users_dir, os.getenv("USER"))

# Utilities

def join_sandbox(path):
	"""Get full path to PATH (must be relative to the Scratchbox
	   sandbox)."""

	if path.startswith("/"):
		return join(sandbox_dir, path[1:])
	else:
		return join(sandbox_dir, path)

def flatten(list, sep = ":"):
	"""Flatten a list of strings into a single string using
	   a separator."""

	s = ""
	if list:
		for i in list:
			if s:
				s = s + sep + i
			else:
				s = i
	return s

def printable(s, none = "none"):
	"""Return S, or None if S is an empty string."""

	if s:
		return s
	else:
		return none

def processable(s, rep = None):
	"""Convert "none" and empty string to REP."""

	if not s or s == "none":
		return rep
	else:
		return s

def symlink(src,dst):
	"""Wrapper script for symlinking, behaves like ln -sf."""
	try:
		os.unlink(dst)
	except OSError:
		pass
	return os.symlink(src,dst)

##
## Error

class Error (Exception):
	def __init__(self, s):
		Exception.__init__(self, s)

class ForceError (Error):
	def __init__(self, s):
		Error.__init__(self, s)

class DownloadError (Error):
	def __init__(self, s):
		Error.__init__(self, s)

class SessionError (Error):
	def __init__(self, s):
		Error.__init__(self, s)

##
## Current target

def current(throw = True):
	"""Get the name of the active target.  Raises and error if there
	   is no current target and THROW is set."""

	path = join_sandbox("/targets/links/scratchbox.config")
	if islink(path):
		conf = basename(os.readlink(path))
		pos = conf.rfind(".")
		if pos > 0:
			return conf[:pos]
	if throw:
		raise Error("No current target")

##
## Setup

def setup(target = None, compiler = None, devkits = None, cputransp = None, force = False):
	"""Creates a target configuration file ("/targets/TARGET.config")
	   or modifies an old one if one exists and FORCE is True."""

	# Target

	if not target:
		target = current(None)
		if not target:
			raise Error("No target name provided and no current target")

	if ' ' in target:
		raise Error('Invalid target name: ' + target)

	# Check input values

	if compiler and compiler not in list_compilers():
		raise Error("No such compiler: " + compiler)

	if devkits:
		devkits = normalize_devkits(devkits, verbose=True)

	if cputransp:
		if not exists(join_sandbox(cputransp)):
			found = False
			alt = join("/scratchbox/tools/bin", cputransp)
			if exists(join_sandbox(alt)):
				cputransp = alt
				found = True
			elif devkits:
				for devkit in devkits:
					alt = join(devkits_dir, devkit + "/bin/" + cputransp)
					if exists(alt):
						cputransp = alt
						found = True
						break

			if not found:
					raise Error("No such CPU-transparency method: '" + cputransp +
								"'.\n\nPerhaps you forgot to select the devkit which includes it.")

	# Fill in old values

	oldconf = read_config(target)
	if oldconf:
		if not force:
			raise ForceError("Target already exists: " + target)

		if compiler == None:
			compiler = oldconf.get("SBOX_CROSS_GCC_NAME")

		if devkits == None:
			devkits = oldconf.get("SBOX_DEVKIT_NAMES", "").split(":")

		if cputransp == None:
			cputransp = oldconf.get("SBOX_CPUTRANSPARENCY_METHOD")

	if not compiler:
		raise Error("No compiler specified")

	# Write config file

	compinfo = get_compiler_info(compiler)

	if cputransp == None:
		cputransp = ""

	def clean_prefix(parts,tag='-',tail='-'):
		return tag.join([part for part in parts if part])+tail

	real_prefix       = clean_prefix([compinfo.subarch, compinfo.os])
	real_subst_prefix = "sbox-%s" % real_prefix

	real_prefix_list = set([real_prefix])

	if compinfo.arch != compinfo.subarch:
		real_prefix_list.add(clean_prefix([compinfo.arch, compinfo.os]))

	#KLUDGE:
	#Provide a toolchain prefix without a vendor name in case of unknown or none
	for vendor in ['unknown-','none-']:
		if vendor in compinfo.os:
			osname = compinfo.os.replace(vendor, '')
			real_prefix_list.add('%s-%s-' % (compinfo.arch, osname))

	#Add '${arch}-linux-' if needed:
	if compinfo.os != 'linux':
		real_prefix_list.add('%s-linux-' % compinfo.arch)

	#Include the most real prefix...
	machine = dump_machine(compiler)
	if machine:
		prefix = machine + '-'
		real_prefix_list.add(prefix)

		for vendor in ('unknown-', 'none-'):
			trimmed_prefix = prefix.replace(vendor, '')
			real_prefix_list.add(trimmed_prefix)

	#Add 'i386-linux-' and 'i486-linux-gnu-' for Debian's benefit
	if re.match('^i.86$', compinfo.arch):
		real_prefix_list.add('i386-linux-')
		real_prefix_list.add('i486-linux-gnu-')

	HOST_LD_ARGS = '"-rpath /scratchbox/compilers/host-gcc/lib/' + \
		       ' -L/scratchbox/compilers/host-gcc/lib"'

	if compiler == "host-gcc":
		setup_script       = ""
		default_prefix     = "host-"
		cross_prefix_list  = real_prefix_list
		cross_subst_prefix = "host-"
		cross_specs        = "host-gcc.specs"
		cross_ld_args      = HOST_LD_ARGS
	else:
		setup_script       = join(compinfo.dir, "target_setup.sh")
		default_prefix     = real_prefix
		cross_prefix_list  = real_prefix_list
		cross_subst_prefix = real_subst_prefix
		cross_specs        = "gcc.specs"
		cross_ld_args      = ""

	path = join_sandbox("/targets/%s.config" % target)
	out = file(path, "w")

	out.write("# Automatically generated configuration data. DO NOT TOUCH!\n")
	out.write("\n")
	out.write("SBOX_TARGET_NAME=%s\n" % target)
	out.write("SBOX_CPU=%s\n" % compinfo.subarch)
	out.write("SBOX_OS=%s\n" % compinfo.os)
	out.write("SBOX_CLIB=%s\n" % compinfo.clib)
	out.write("SBOX_CPUTRANSPARENCY_METHOD=%s\n" % cputransp)
	out.write("SBOX_DEVKIT_NAMES=%s\n" % flatten(devkits))
	out.write("\n")
	out.write("# DO NOT USE THESE IN YOUR MAKEFILES\n")
	out.write("\n")
	out.write("SBOX_DEFAULT_GCC_PREFIX=%s\n" % default_prefix)
	out.write("SBOX_TARGET_DIR=/targets/%s\n" % target)
	out.write("SBOX_TARGET_SETUP_SCRIPT=%s\n" % setup_script)
	out.write("\n")
	out.write("SBOX_CROSS_GCC_NAME=%s\n" % compinfo.name)
	out.write("SBOX_CROSS_GCC_PREFIX_LIST=%s\n" % flatten(cross_prefix_list))
	out.write("SBOX_CROSS_GCC_SUBST_PREFIX=%s\n" % cross_subst_prefix)
	out.write("SBOX_CROSS_GCC_SPECS_FILE=%s/%s\n" % (compinfo.dir, cross_specs))
	out.write("SBOX_CROSS_GCC_DIR=%s/bin\n" % compinfo.dir)
	out.write("SBOX_CROSS_GCC_LD_ARGS=%s\n" % cross_ld_args)
	out.write("\n")
	out.write("SBOX_HOST_GCC_NAME=host-gcc\n")
	out.write("SBOX_HOST_GCC_PREFIX_LIST=host-\n")
	out.write("SBOX_HOST_GCC_SUBST_PREFIX=host-\n")
	out.write("SBOX_HOST_GCC_SPECS_FILE=/scratchbox/compilers/host-gcc/host-gcc.specs\n")
	out.write("SBOX_HOST_GCC_DIR=/scratchbox/compilers/host-gcc/bin\n")
	out.write("SBOX_HOST_GCC_LD_ARGS=%s\n" % HOST_LD_ARGS)

##
## Install

def install(target = None, clibrary = False, cruntime = False, etc = False, devkits = False, links = False, tools = []):
	"""Installs the specified groups of files to TARGET's filesystem
	   directory."""

	# Target

	if not target:
		target = current()

	# Config

	conf = read_config(target)
	if not conf:
		raise Error("No such target: " + target)

	compiler = conf["SBOX_CROSS_GCC_NAME"]
	compinfo = get_compiler_info(compiler)
	root = conf["SBOX_TARGET_DIR"]

	root = join_sandbox(root)
	if not exists(root):
		os.makedirs(root)

	# C-library

	if clibrary or cruntime:
		extra = ""
		if cruntime:
			extra = "cruntime"
		script = conf.get("SBOX_TARGET_SETUP_SCRIPT")
		if not compinfo.prefixedInstall and target != current(None):
			raise Error("Cowardly refusing to install C-library to %s with a script that "%target +
						"probably breaks the current target.\nSelect target `%s' first."%target)
		if script and exists(script) and not run_script(script, root, extra):
			raise Error("Failed to install C-library to target: " + target)

	# /etc

	if etc:
		targetdir = join(root, "etc")
		if not exists(targetdir):
			os.makedirs(targetdir)

		basedir = join(etc_dir, "target_etc_base")
		for i in os.listdir(basedir):
			shutil.copy(join(basedir, i), targetdir)

		# mtab

		mtab = join(targetdir, "mtab")
		if exists(mtab):
			os.remove(mtab)
		os.symlink("/proc/mounts", mtab)

		# passwd

		path = join(targetdir, "passwd")
		existed = exists(path)

		out = file(path, "a")

		if not existed:
			out.write("root:x:0:0:root:/root:/bin/sh\n")

		out.write(flatten([os.getenv("USER"), "x", str(os.getuid()), str(os.getgid()),
				   "", os.getenv("HOME"), "/bin/sh"]))
		out.write("\n")

		# group

		path = join(targetdir, "group")
		existed = exists(path)

		out = file(path, "a")

		if not existed:
			out.write("root:*:0:\n")

		gid = os.getgid()

		groupname = os.getenv("_SBOX_USER_GROUPNAME")
		if not groupname:
			groupname = grp.getgrgid(gid)[0]

		out.write(flatten([groupname, "x", str(gid)]))
		out.write("\n")

	# Devkits

	if devkits:
		for name in conf.get("SBOX_DEVKIT_NAMES", "").split(":"):
			script = join(devkits_dir, name, "target_setup.sh")
			if exists(script) and not run_script(script, root):
				raise Error("Failed to install " + name + \
						" devkit to target: " + target)

	# Links

	if links:
		targetdir = join(root, "usr/lib/debug")
		dbg_dir = '/usr/lib/debug'
		if os.system('mkdir -p %s/targets' % targetdir):
			raise Error("Failed to install debug links to target: " + target)
		if symlink(dbg_dir, '%s/targets/%s'%(targetdir, target)):
			raise Error("Failed to install debug links to target: " + target)


	# Tools

	root_usr = join(root, "usr")
	for name in tools:
		copy_device_tool(name, compiler, root_usr)

def run_script(script, root, extra = ""):
	"""Execute SCRIPT using shell.  ROOT is passed as the first
	   argument. EXTRA is passed as the second argument, if
	   it exists."""
	sh=join(tools_dir, "bash")
	#Check for script type
	line=open(script).readline(128)
	if line.startswith('#!'):
		return not os.spawnv(os.P_WAIT, script, [script, root, extra])
	else:
		return not os.spawnv(os.P_WAIT, sh, [sh, script, root, extra])

def copy_device_tool(name, compiler, destroot):
	"""Copy a device tool to directory DESTROOT."""

	versions = find_device_tool_versions(name, compiler)
	if not versions:
		raise Error("No %s versions found for compiler: %s" % (name, compiler))

	version, path = versions[-1]
	print "Installing %s version %s..." % (name, version)
	copy_hierarchy(path, destroot)

def find_device_tool_versions(name, compiler):
	dict = {}

	if exists(devicetools_dir):
		for program in os.listdir(devicetools_dir):
			if program.startswith(name + "-"):
				path = join(devicetools_dir, program, compiler)
				if exists(path):
					version = program[len(name)+1:]
					numerical = parse_version(version)
					dict[numerical] = (version, path)

	list = dict.keys()
	list.sort()

	return [dict[i] for i in list]

def copy_hierarchy(srcroot, destroot):
	"""Copy directory hierarchy."""

	prefixlen = len(srcroot) + 1

	for info in os.walk(srcroot):
		srcdir = info[0]

		destdir = join(destroot, srcdir[prefixlen:])

		if not islink(srcdir):
			if islink(destdir):
				os.remove(destdir)
			if not exists(destdir):
				os.makedirs(destdir)

		# symlinks
		for i in info[1]:
			src = join(srcdir, i)
			dest = join(destdir, i)
			if islink(src):
				if exists(dest):
					if islink(dest):
						os.remove(dest)
					else:
						shutil.rmtree(dest)
				os.symlink(os.readlink(src), dest)

		# regular files
		for i in info[2]:
			src = join(srcdir, i)
			dest = join(destdir, i)

			if exists(dest):
				os.remove(dest)
			if islink(src):
				src_real = os.readlink(src)
				# avoid an exception with symlinks pointing to themselves
				if basename(src) != src_real:
					os.symlink(src_real, dest)
			else:
				shutil.copy(src, dest)

def parse_version(s):
	def parse(s):
		num = 0
		tail = None
		for i in xrange(len(s)):
			c = s[i]
			if c.isdigit():
				num = num * 10 + int(c)
			else:
				tail = s[i:]
				break
		if str:
			return (num, tail)
		else:
			return (num,)

	return tuple([parse(i) for i in s.split('.')])

##
## Rootstrap

def rootstrap(target = None, path = None, url = None):
	"""Extract a rootstrap (tarball) to TARGET's filesystem directory.
	   If PATH is specified, it is extracted; if URL is specified, it
	   is first downloaded to /tmp using wget and then extracted."""

	# Target

	if not target:
		target = current()

	# Target dir

	conf = read_config(target)
	if not conf:
		raise Error("No such target: " + target)

	root = conf["SBOX_TARGET_DIR"]

	root = join_sandbox(root)
	if not exists(root):
		os.makedirs(root)

	# Download

	if url:
		base = basename(url)
		if not base:
			raise DownloadError("Invalid rootstrap URL: " + url)

		tmp = join("/tmp", "sb-config-" + os.getenv("USER"))
		if not exists(tmp):
			os.makedirs(tmp)

		path = join(tmp, base)
		if exists(path):
			os.remove(path)

		bin = join(tools_dir, "wget")
		if os.spawnv(os.P_WAIT, bin, [bin, "-O", path, url]):
			if exists(path):
				os.remove(path)
			raise DownloadError("Failed to download rootstrap: " + url)

		if not exists(path):
			raise DownloadError("Failed to locate downloaded rootstrap: " + path)

	# Extract

	bin = join(tools_dir, "tar")
	print "Unpacking rootstrap..."
	if os.spawnv(os.P_WAIT, bin, [bin, "xfz", path, "-C", root, "--exclude", "./dev/*"]):
		raise Error("Failed to extract rootstrap: " + path)

	if url:
		os.remove(path)

##
## Select

def select(target = None):
	"""Make TARGET active."""

	# Target

	if not target:
		target = current()

	reselecting = (current(None) == target)


	# Config

	conf = read_config(target)
	if not conf:
		raise Error("No such target: " + target)

	root = conf["SBOX_TARGET_DIR"]

	root = join_sandbox(root)
	if not exists(root):
		os.makedirs(root)

	compdir = join("/scratchbox/compilers/", conf["SBOX_CROSS_GCC_NAME"])

	# Check login sessions

	sessions = list_sessions()

	maximum = 0
	for tty in sessions:
		if os.getpid() in sessions[tty]:
			maximum = 1
			break

	if len(sessions) > maximum:
		if maximum > 0:
			word = "other "
		else:
			word = ""
		if reselecting:
			print "Note: Couldn't reselect the target. There are other Scratchbox sessions open."
			return
		else:
			raise SessionError("You must close your " + word + "Scratchbox sessions first")

	# Create links

	linkdir = join_sandbox("/targets/links")
	conflink = join(linkdir, "scratchbox.config")
	have_previous = islink(conflink)

	if have_previous:
		from sb import sbrsh
		sbrsh.umount(target)
		del sbrsh

		os.remove(conflink)

	if not exists(linkdir):
		os.makedirs(linkdir)

	os.symlink("/targets/%s.config" % target, conflink)

	for link in open(join(etc_dir, "rootdirs")).read().split():
		dirpath = join("/targets", target, link)

		physical_dirpath = join_sandbox(dirpath)
		if not islink(physical_dirpath):
			if not exists(physical_dirpath):
				os.makedirs(physical_dirpath)

		linkpath = join(linkdir, link)
		if islink(linkpath):
			os.remove(linkpath)

		os.symlink(dirpath, linkpath)

	archlink = join(linkdir, "arch_tools")
	if islink(archlink):
		os.remove(archlink)
	archdest = join(compdir, "arch_tools")
	if not exists(archdest) and exists(join(compdir,'libtool')):
		archdest=join(compdir,'libtool')

	os.symlink(archdest, archlink)

	# Restart

	pid = os.getenv("_SBOX_SHELL_PID")
	if pid:
		restart = os.getenv("_SBOX_RESTART_FILE")
		if restart:
			file(restart, "w").close()
			os.kill(int(pid), signal.SIGHUP)

##
## killall


def killall(signal=15):
	def parents(pid):
		try:
			f = open(join('/proc',str(pid),'stat'))
			parent=int(f.readlines()[0].split(' ')[3])
			f.close()
			if parent != pid:
				return [parent]+parents(parent)
			else:
				return []
		except IOError:
			return []
	protected=dict.fromkeys([os.getpid()]+parents(os.getpid()))
	for pid in [x for l in list_sessions().itervalues() for x in l if x not in protected]:
		try:
			os.kill(pid,signal)
		except OSError:
			continue



##
## Remove, Reset

def read_target_root(target):
	"""Get TARGET's root directory."""

	conf = read_config(target)
	if conf:
		return join_sandbox(conf["SBOX_TARGET_DIR"])
	else:
		return None

def remove(target = None, force = False):
	"""Delete TARGET's configuration (including sbrsh configuration)
	   and filesystem directory.  If the target directory exists but
	   FORCE is not True, an exception is raised.  The active target
	   cannot be removed."""

	if not target:
		target = current()

	root = read_target_root(target)
	if not root:
		if force:
			return
		raise Error("No such target: " + target)

	if not force:
		raise ForceError("You have to force me to remove a target...")

	if exists(root):
		if islink(root):
			os.remove(root)
		else:
			remove_contents(root)
			os.rmdir(root)

	sbrsh = join_sandbox("/targets/%s.sbrsh" % target)
	if exists(sbrsh):
		os.remove(sbrsh)

	os.remove(join_sandbox("/targets/%s.config" % target))

	if exists(join_sandbox("/targets/%s.environment" % target)):
		os.remove(join_sandbox("/targets/%s.environment" % target))

def reset(target = None, force = False):
	"""Deletes the contents of TARGET's filesystem directory.  If the
	   target directory exists but FORCE is not True, an exception is
	   raised.  If TARGET is currently active, the directory will be
	   initialised so that it can be used without reselecting it."""

	if not target:
		target = current()

	root = read_target_root(target)
	if not root:
		if force:
			return
		raise Error("No such target: " + target)

	if not exists(root):
		return

	if not force:
		raise ForceError("You have to force me to reset a target...")

	remove_contents(root)

	# Create root directories so that links work
	if target == current(None):
		for link in open(join(etc_dir, "rootdirs")).read().split():
			dirpath = join("/targets", target, link)

			physical_dirpath = join_sandbox(dirpath)
			if not exists(physical_dirpath):
				os.makedirs(physical_dirpath)

def remove_contents(root):
	for path, dirs, files in os.walk(root, False):
		for i in files:
			os.remove(join(path, i))

		for i in dirs:
			j = join(path, i)

			if islink(j):
				os.remove(j)
			else:
				os.rmdir(j)

##
## Config Info

class ConfigInfo:
	pass

def read_config(target = None):
	"""Parses TARGET's configuration file into a map."""

	if not target:
		target = current()

	path = join_sandbox("/targets/%s.config" % target)
	if isfile(path):
		config = {}
		for line in file(path).readlines():
			pos = line.find("#")
			if pos >= 0:
				line = line[:pos]
			if line.find("=") >= 0:
				key, val = line.split("=", 1)
				config[key.strip()] = val.strip()
		return config

def read_compiler(compiler):
	"""Parses COMPILER's "compiler-name" file into a list."""

	path = join(compilers_dir, compiler, "compiler-name")
	if isfile(path):
		return file(path).readline().strip().split(":")

def get_target_info(target = None):
	"""Returns a data object describing TARGET.  The object contains
	   the following fields: compiler (string), devkits (string list)
	   and cputransp (string)."""

	if not target:
		target = current()

	conf = read_config(target)
	if not conf:
		return None

	info = ConfigInfo()

	try:
		info.compiler = conf["SBOX_CROSS_GCC_NAME"]
	except KeyError:
		return None

	dk = conf.get("SBOX_DEVKIT_NAMES", "").strip().split(":")
	if len(dk) == 1 and not dk[0]:
		dk = []

	info.devkits = dk

	ct = conf.get("SBOX_CPUTRANSPARENCY_METHOD", "").strip()
	if ct:
		if ct.startswith("/scratchbox/tools/bin/"):
			ct = ct[len("/scratchbox/tools/bin/"):]

	info.cputransp = ct

	return info

def get_compiler_info(compiler = None):
	"""Returns a data object describing COMPILER.  The object contains
	   the following strings: name, dir, arch, subarch, os, clib."""

	if not compiler:
		compiler = get_target_info().compiler
		if not compiler:
			return None

	try:
		conf = read_compiler(compiler)

		info = ConfigInfo()
		info.name    = conf[0]
		info.dir     = join("/scratchbox/compilers", info.name)
		info.subarch = conf[2]
		info.os      = conf[3]
		info.clib    = conf[4]

		if len(conf) > 5:
			info.arch = conf[5]
		else:
			info.arch = info.subarch

		def check_compiler_targetsetup():
			"""Check the target_setup.sh script.
			Returns true if it is ok to install c-library with some prefix"""
			path = join(compilers_dir, compiler, "target_setup.sh")
			setupFile=None
			if not isfile(path):
				return True
			try:
				setupFile=open(path,'r')
				if 'python' in setupFile.readline():
					return True
				for line in setupFile:
					if "$1" in line:
						return True
				return False
			finally:
				setupFile.close()

		info.prefixedInstall=check_compiler_targetsetup()

		return info

	except:
		return None

def dump_machine(compiler):
	info = get_compiler_info(compiler)
	file = '%s-%s-gcc' % (info.subarch, info.os)
	path = join(compilers_dir, compiler, 'bin', file)
	args = [path, '-dumpmachine']
	stdin, stdout = os.popen2(args)
	line = stdout.readline()
	if line:
		return line[:-1]
	else:
		return None

##
## List

def list_paths(dir):
	"""Returns full paths of all non-hidden files directory DIR."""

	list = []
	for file in os.listdir(dir):
		if file[0] != ".":
			list.append(join(dir, file))
	return list

def list_compilers():
	"""Returns all available compilers."""

	list = []
	if exists(compilers_dir):
		for name in os.listdir(compilers_dir):
			if get_compiler_info(name):
				list.append(name)
	list.sort()
	return list

def list_devkits():
	"""Returns all available devkits."""

	list = []
	if exists(devkits_dir):
		for dir in list_paths(devkits_dir):
			if isdir(dir):
				list.append(basename(dir))
	list.sort()
	return list

def get_devkit_deps(devkit):
	"""Returns dependencies (devkit names) of DEVKIT."""

	depends = []

	path = join(devkits_dir, devkit, "depends")
	if exists(path):
		depends = file(path).read().split()

	available = set(list_devkits())
	for dep in depends:
		if dep not in available:
			print >>sys.stderr, 'Warning: Devkit', devkit, \
			      'depends on non-existent devkit:', dep

	return depends

def normalize_devkits(devkits, verbose=False):
	available = set(list_devkits())
	for name in devkits:
		if name not in available:
			raise Error("No such devkit: " + name)

	while True:
		new_devkits = set()

		for name in devkits:
			for dep in get_devkit_deps(name):
				if dep not in devkits:
					if verbose:
						print "Adding unselected devkit:", dep
					new_devkits.add(dep)

		if not new_devkits:
			break

		devkits = list(new_devkits) + devkits

	depends = set()
	for name in devkits:
		for dep in get_devkit_deps(name):
			depends.add((name, dep))

	return reverse_topological_sort(devkits, depends)

def list_cputransps(devkits=[]):
	"""Returns all standard CPU-transparency methods (specified in
	   /scratchbox(|devkits/devkit)/etc/cputransp-methods)."""

	path = join(etc_dir, "cputransp-methods")
	methods = file(path).read().split()

	for devkit in devkits:
		path = join(devkits_dir, devkit + "/etc/cputransp-methods")
		if isfile(path):
			methods += file(path).read().split()
	return methods

def list_targets():
	"""Retuns all available targets."""

	list = []
	for path in list_paths(join_sandbox("/targets")):
		if isfile(path) and path.endswith(".config"):
			conf = basename(path)
			list.append(conf[:conf.rindex(".")])
	list.sort()
	return list

def list_sessions():
	"""Returns all terminal devices currently in use within the
	   user's Scratchbox sandbox."""

	map = {}
	inode = os.stat(sandbox_dir).st_ino

	for pidstr in os.listdir("/proc"):
		try:
			pid = int(pidstr)
		except ValueError:
			continue

		path = join("/proc", pidstr, "root")
		if not islink(path) or not os.access(path, os.R_OK):
			continue

		if os.stat(path).st_ino != inode:
			continue

		for fd in xrange(3):
			link = join("/proc", pidstr, "fd", str(fd))
			if not islink(link):
				continue

			tty = os.readlink(link)
			if not tty.startswith("/dev/"):
				continue

			if tty not in map:
				map[tty] = []
			map[tty].append(pid)
			break

	return map

##
## Version

def version():
	"""Returns the Scratchbox version string."""

	path = join(etc_dir, "scratchbox-version")
	return file(path).readline().strip()

##
## Graph sort

def reverse_topological_sort(vertices, edges):
	def get_out_degree(vertex):
		value = 0
		for head, tail in edges:
			if vertex == head:
				value += 1
		return value

	def get_in_edges(vertex):
		list = []
		for edge in edges:
			head, tail = edge
			if vertex == tail:
				list.append(edge)
		return list

	queue = []
	out_degrees = {}

	for vertex in vertices:
		out_degree = get_out_degree(vertex)
		if out_degree == 0:
			queue.append(vertex)
		else:
			out_degrees[vertex] = out_degree

	list = []

	while queue:
		vertex = queue.pop(0)
		list.append(vertex)

		for edge in get_in_edges(vertex):
			head, tail = edge
			out_degrees[head] -= 1
			if out_degrees[head] == 0:
				queue.append(head)

	if len(list) != len(vertices):
		raise Error('Cyclic devkit dependencies')

	return list
